/* mconfig (c)1996 University of Michigan 
 * By Dave Thaler <thalerd@eecs.umich.edu> 
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <ctype.h>
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifndef HAVE_GETHOSTNAME
#include <sys/systeminfo.h>
#endif
#include <netdb.h>
#include <arpa/inet.h>
#include <tcl.h>
#include <tk.h>
#include "snmp.h"
#include "../snmplib/snmp.h"
#include "../snmplib/asn1.h"
#include "../snmplib/snmp_impl.h"
#include "../snmplib/snmp_api.h"
#include "../snmplib/snmp_client.h"
#include "../snmplib/view.h"
#include "../snmplib/acl.h"
#include "../snmplib/mib.h"
#include "../snmplib/parse.h"
#include "util.h"
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif

char *snmp_errlist[]={ "",
   "Too Big", "No Such Value", "Bad value", "Read Only",
   "Gen Err", "No Access", "Wrong Type", "Wrong Length",
   "Wrong Encoding", "Wrong Value", "No Creation", "Inconsistent Value",
   "Resource Unavailable", "Commit Failed", "Undo Failed", 
   "Authorization Error", "Not Writable"
};

void
show_error(err)
   int err;
{
   printf("ERROR: %s\n", snmp_errlist[err]);
}

int  /* error code */
snmpset(ss, names, values, num_names)
   struct snmp_session *ss;
   char **names;
   char **values;
   int    num_names;
{
   struct snmp_pdu *pdu, *response = NULL;
   int count;
   int name_length;
   oid name[MAX_NAME_LEN];
   int status, err;
   struct variable_list *vpc;

   pdu = snmp_pdu_create(SET_REQ_MSG);

   /* Convert names to OIDs */
   for(count = 0; count < num_names; count++){
      name_length = MAX_NAME_LEN;
      if (!get_objid(names[count], name, &name_length))
         printf("Invalid object identifier: %s\n", names[count]);
      snmp_add_null_var(pdu, name, name_length);
   }

   /* Fill in values */
   for (count = 0; count < num_names; count++) {
      vpc = vpseek(pdu->variables,count);
      if (!isdigit(values[count][0])) {        /* DisplayString */
         vpc->type = ASN_OCTET_STR;
         vpc->val_len = strlen(values[count]);
         vpc->val.string = (u_char*)malloc(vpc->val_len);
         memcpy(vpc->val.string, values[count], vpc->val_len);
      } else if (strchr(values[count], '.')) { /* IpAddress */
         vpc->type = IPADDRESS; /* write_vif expects str LocalAddress,etc */
         vpc->val_len = sizeof(int32);
         vpc->val.string = (u_char*)malloc(vpc->val_len);
         *vpc->val.integer = inet_addr(values[count]);
      } else {                                 /* INTEGER */
         vpc->type = ASN_INTEGER;
         vpc->val_len = 4;
         vpc->val.string = (u_char*)malloc(vpc->val_len);
         *vpc->val.integer = atoi(values[count]);
      }
   }

   err = 0;
   for (;;) {
      status = snmp_synch_response(ss, pdu, &response);
      if (status == STAT_SUCCESS) {
         if (response->errstat!=SNMP_ERR_NOERROR)
            show_error(response->errstat);
         err = response->errstat;
      } else if (status == STAT_TIMEOUT) {
         printf("No Response from host\n");
      } else  {    /* status == STAT_ERROR */
         printf("An error occurred, Quitting\n");
      }
      break;
   }

   if (response)
      snmp_free_pdu(response);
   return err;
}

#define sysDescr         ".1.3.6.1.2.1.1.1.0"
#define sysUpTime        ".1.3.6.1.2.1.1.3.0"
#define sysContact       ".1.3.6.1.2.1.1.4.0"
#define sysLocation      ".1.3.6.1.2.1.1.6.0"
typedef struct {
   char *name;
   char *oid;
} system_t;

system_t system_vars[]={
 { "Hostname",     ".1.3.6.1.2.1.1.5.0", },
 { "Version",      ".1.3.6.1.3.62.1.1.1.0", },
 { "Description",  sysDescr, },
 { "Location",     sysLocation, },
 { "Contact",      sysContact, },
 { NULL, NULL },
};

char *
snmpget_string(ss, obj, err, tm)
   struct snmp_session *ss;
   char *obj;
   int  *err;
   int  *tm;
{
   struct snmp_pdu *pdu;
static char buff[4096];
   char tmbuff[256];
   char *vars[3];

   if (!ss) return NULL;
   vars[0] = obj;
   vars[1] = sysUpTime;
   pdu = snmpget(ss, GET_REQ_MSG, vars, 2, err);
   if (!pdu) return NULL;
   tmbuff[0]=buff[0]='\0';
   accumulate_value(buff, pdu->variables);
   accumulate_counter(tmbuff, pdu->variables->next_variable);
   if (tm) 
      *tm = atoi(tmbuff);
   snmp_free_pdu(pdu);
   return buff;
}

char *vifTclVars[] = {
"vifLocal($I(-1))", "$0",
"vifMetric($I(-1))", "$3",
"vifThresh($I(-1))", "$4",
"vifRate($I(-1))", "$5",
"vifType($I(-1))", "$6",
"vifOperState($I(-1))", "$7",
"vifAdminState($I(-1))", "$8",
"vifRemote($I(-1))", "$S(1,2)",
"vifState($I(-1))", "$D(6,7)",
NULL, NULL
};

char *neighborTclVars[]={
NULL, NULL
};
char *boundaryTclVars[]={
"boundaryVif($R)",  "$I(-9)", 
"boundaryAddr($R)", "$A(-8)", 
"boundaryMask($R)", "$A(-4)",
NULL, NULL
};
char *altnetTclVars[]={
"altnetVif($R)",  "$I(-9)", 
"altnetAddr($R)", "$A(-8)", 
"altnetMask($R)", "$A(-4)",
NULL, NULL
};

extern Tcl_Interp *tcl;

const int tclFlag = 1;

void
output_header(str)
   char *str;
{
   char buff[512];

   while (str[ strlen(str)-1 ]=='\n')
      str[ strlen(str)-1 ]='\0';
   sprintf(buff, "outputHeader {%s}", str);
   Tcl_Eval(tcl, buff);
}

void
setAnswers(vp, cumul, oldcumul,nvars,row)
   struct variable_list *vp;
   char cumul[10][256];
   char oldcumul[10][256];
   int nvars;
   int row;
{
   int j;
   char s1[256], s2[256];

   /* Save data */
   switch(table) {
   case TAB_VIFS:
      for (j=0; vifTclVars[j]; j+=2) {
         expand_str(vifTclVars[j],  vp,cumul,oldcumul,nvars,row,s1);
         expand_str(vifTclVars[j+1],vp,cumul,oldcumul,nvars,row,s2);
         Tcl_SetVar(tcl,s1,s2,TCL_GLOBAL_ONLY);
      }
      break;

   case TAB_NEIGHBORS:
      for (j=0; neighborTclVars[j]; j+=2) {
         expand_str(neighborTclVars[j],  vp,cumul,oldcumul,nvars,row,s1);
         expand_str(neighborTclVars[j+1],vp,cumul,oldcumul,nvars,row,s2);
         Tcl_SetVar(tcl,s1,s2,TCL_GLOBAL_ONLY);
      }
      break;

   case TAB_BOUNDARIES:
      for (j=0; boundaryTclVars[j]; j+=2) {
         expand_str(boundaryTclVars[j],  vp,cumul,oldcumul,nvars,row,s1);
         expand_str(boundaryTclVars[j+1],vp,cumul,oldcumul,nvars,row,s2);
         Tcl_SetVar(tcl,s1,s2,TCL_GLOBAL_ONLY);
      }
      break;

   case TAB_ALTNETS:
      for (j=0; altnetTclVars[j]; j+=2) {
         expand_str(altnetTclVars[j],  vp,cumul,oldcumul,nvars,row,s1);
         expand_str(altnetTclVars[j+1],vp,cumul,oldcumul,nvars,row,s2);
         Tcl_SetVar(tcl,s1,s2,TCL_GLOBAL_ONLY);
      }
      break;

   }
}

void
output_row(str)
   char *str;
{
   char buff[512];

   sprintf(buff, "outputRow {%s}", str);
   Tcl_Eval(tcl, buff);

}

int
getVifs(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   table = TAB_VIFS;
   display_table(ss);
   return TCL_OK;
}

int
getSystem(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *str;
   char buff[4096], buff2[80];
   int err;
   int i;
   int tm;

   for (i=0; system_vars[i].name; i++) {
      str = snmpget_string(ss, system_vars[i].oid, &err, &tm);
      sprintf(buff, ".f.list insert end {%-15s %s}", 
       system_vars[i].name, str);
      Tcl_Eval(tcl, buff);
      Tcl_SetVar(tcl, system_vars[i].name, str, TCL_GLOBAL_ONLY);
   }
   sprintf(buff, ".f.list insert end {%-15s %s}\n", "Uptime",
    uptimeString(tm,buff2));
   Tcl_Eval(tcl, buff);

   return TCL_OK;
}

int
getNeighbors(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   table = TAB_NEIGHBORS;
   display_table(ss);
   return TCL_OK;
}

int
getBoundaries(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   table = TAB_BOUNDARIES;
   display_table(ss);
   return TCL_OK;
}

int
getAltNets(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   table = TAB_ALTNETS;
   display_table(ss);
   return TCL_OK;
}

int
openSession(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   if (argc>1) {
      if (ss)
         snmp_close(ss);
      ss = open_session(argv[1],dest_port,community,retransmission,timeout);
   }
   return TCL_OK;
}

/* syntax: modifySystem description location contact */
int
modifySystem(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[3], *values[3];
   char name[3][256], value[3][256];
   int i, num=0;

   for (i=0; i<3; i++) {
      names[i]  = name[i];
      values[i] = value[i];
   }

   if (argc>3) {
      int err;

      if (strcmp(argv[1],Tcl_GetVar(tcl,"Description",TCL_GLOBAL_ONLY))){
         strcpy(name[num],  sysLocation);
         strcpy(value[num], argv[2]);
         num++;
      }
      if (strcmp(argv[2],Tcl_GetVar(tcl,"Location",TCL_GLOBAL_ONLY))){
         strcpy(name[num],  sysLocation);
         strcpy(value[num], argv[2]);
         num++;
      }
      if (strcmp(argv[3],Tcl_GetVar(tcl,"Contact",TCL_GLOBAL_ONLY))){
         strcpy(name[num],  sysContact);
         strcpy(value[num], argv[3]);
         num++;
      }
      if (num)
         err = snmpset(ss, names, values, num);
   }
   return TCL_OK;
}

/* syntax: modifyVif vif local remote metric thresh rate adminstate */
int
modifyVif(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[7], *values[7];
   char name[7][256], value[7][256];
   int i, num=0;

   for (i=0; i<7; i++) {
      names[i]  = name[i];
      values[i] = value[i];
   }

   if (argc>7) {
      int err;
      int vifi = atoi(argv[1]);

      if (strcmp(argv[2],Tcl_GetVar2(tcl,"vifLocal",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceLocalAddress, vifi);
         strcpy(value[num], argv[2]);
         num++;
      }
      if (strcmp(argv[3],Tcl_GetVar2(tcl,"vifRemote",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceRemoteAddress, vifi);
         /* XXX convert hostname to ipaddress */
         strcpy(value[num], argv[3]);
         num++;
      }
      if (strcmp(argv[4],Tcl_GetVar2(tcl,"vifMetric",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceMetric, vifi);
         strcpy(value[num], argv[4]);
         num++;
      }
      if (strcmp(argv[5],Tcl_GetVar2(tcl,"vifThresh",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceThresh, vifi);
         strcpy(value[num], argv[5]);
         num++;
      }
      if (strcmp(argv[6],Tcl_GetVar2(tcl,"vifRate",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceRateLimit, vifi);
         strcpy(value[num], argv[6]);
         num++;
      }
      if (strcmp(argv[7],Tcl_GetVar2(tcl,"vifAdminState",argv[1],TCL_GLOBAL_ONLY))){
         sprintf(name[num], "%s.%d", dvmrpVInterfaceStatus, vifi);
         if (!strcmp(argv[7], "active")) {
            sprintf(value[num], "%d", ENTRY_ACTIVE);
            num++;
         } else if (!strcmp(argv[7], "notInService")) {
            sprintf(value[num], "%d", ENTRY_NOTINSERVICE);
            num++;
         }
      }

      if (num)
         err = snmpset(ss, names, values, num);
   }
   return TCL_OK;
}

int
createVif(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>1) {
      int err;

      /* Set dvmrpVInterfaceStatus.idx = create */
      sprintf(name, "%s.%d", dvmrpVInterfaceStatus, atoi(argv[1]));
      names[0]  = name;
      sprintf(value, "%d", ENTRY_CREATEANDWAIT);
      values[0] = value;
      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

int
createBoundary(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>3) {
      int err;
      u_int32 addr = inet_addr(argv[2]);

      if ((addr & 0xff000000) != 0xef000000) { /* scoped range */
printf("Illegal boundary address\n");
         return TCL_OK;
      }

      /* Set dvmrpBoundaryStatus.vif.addr.mask = create */
      sprintf(name, "%s.%s.%s.%s", dvmrpBoundaryStatus, argv[1], argv[2], argv[3]);
      names[0]  = name;
      sprintf(value, "%d", ENTRY_CREATEANDGO);
      values[0] = value;

      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

int
createAltNet(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>3) {
      int err;

      /* Set dvmrpAltNetStatus.vif.addr.mask = create */
      sprintf(name, "%s.%s.%s.%s", dvmrpAltNetStatus, argv[1], argv[2], argv[3]);
      names[0]  = name;
      sprintf(value, "%d", ENTRY_CREATEANDGO);
      values[0] = value;

      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

int
deleteVif(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>1 && strlen(argv[1])) {
      int err;

      /* Set dvmrpVInterfaceStatus.idx = destroy */
      sprintf(name, "%s.%d", dvmrpVInterfaceStatus, atoi(argv[1]));
      names[0]  = name;
      sprintf(value, "%d", ENTRY_DESTROY);
      values[0] = value;
      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

int
deleteBoundary(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>1 && strlen(argv[1])) {
      int err;

      /* Set dvmrpBoundaryStatus.vif.addr.mask = destroy */
      sprintf(name, "%s.%s.%s.%s", dvmrpBoundaryStatus, 
        Tcl_GetVar2(tcl, "boundaryVif",  argv[1], TCL_GLOBAL_ONLY),
        Tcl_GetVar2(tcl, "boundaryAddr", argv[1], TCL_GLOBAL_ONLY),
        Tcl_GetVar2(tcl, "boundaryMask", argv[1], TCL_GLOBAL_ONLY));
      names[0]  = name;
      sprintf(value, "%d", ENTRY_DESTROY);
      values[0] = value;
      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

int
deleteAltNet(data, tcl, argc, argv)
   ClientData data;
   Tcl_Interp *tcl;
   int argc;
   char **argv;
{
   char *names[2], *values[2];
   char name[256], value[256];

   if (argc>1 && strlen(argv[1])) {
      int err;

      /* Set dvmrpAltNetStatus.vif.addr.mask = destroy */
      sprintf(name, "%s.%s.%s.%s", dvmrpAltNetStatus, 
        Tcl_GetVar2(tcl, "altnetVif",  argv[1], TCL_GLOBAL_ONLY),
        Tcl_GetVar2(tcl, "altnetAddr", argv[1], TCL_GLOBAL_ONLY),
        Tcl_GetVar2(tcl, "altnetMask", argv[1], TCL_GLOBAL_ONLY));
      names[0]  = name;
      sprintf(value, "%d", ENTRY_DESTROY);
      values[0] = value;
      err = snmpset(ss, names, values, 1);
   }
   return TCL_OK;
}

typedef struct {
   char  *text;
   int  (*func)();
} command_t;

command_t command[]={
{"getSystem",      getSystem,      },
{"getVifs",        getVifs,        },
{"getNeighbors",   getNeighbors,   },
{"getBoundaries",  getBoundaries,  },
{"getAltNets",     getAltNets,     },
{"openSession",    openSession,    },
{"modifySystem",   modifySystem,   },
{"modifyVif",      modifyVif,      },
{"createVif",      createVif,      },
{"deleteVif",      deleteVif,      },
{"createBoundary", createBoundary, },
{"deleteBoundary", deleteBoundary, },
{"createAltNet",   createAltNet,   },
{"deleteAltNet",   deleteAltNet,   },
{NULL,             NULL            },
};

char *host;

int
App_Init(interp)
   Tcl_Interp *interp;
{
   int i;

   for (i=0; command[i].func; i++)
      Tcl_CreateCommand(tcl, command[i].text, command[i].func, 0, 0);

   Tcl_LinkVar(tcl, "addronly", (char*)&addronly, TCL_LINK_INT );
   Tcl_LinkVar(tcl, "port", (char*)&dest_port, TCL_LINK_INT );
   Tcl_LinkVar(tcl, "retransmission", (char*)&retransmission, TCL_LINK_INT ); 
   Tcl_LinkVar(tcl, "timeout", (char*)&timeout, TCL_LINK_INT ); 

   Tcl_LinkVar(tcl, "community", (char*)&community, TCL_LINK_STRING );
   Tcl_LinkVar(tcl, "host", (char*)&host, TCL_LINK_STRING );
   Tcl_SetVar(tcl, "host", Tcl_GetVar(tcl, "argv", TCL_GLOBAL_ONLY), TCL_GLOBAL_ONLY);
   if (!host || !strlen(host)) {
      char hostname[80];
#ifndef HAVE_GETHOSTNAME
      sysinfo(SI_HOSTNAME, hostname, 80);
#else
      gethostname(hostname,80);
#endif
      Tcl_SetVar(tcl, "host", hostname, 0);
   }
   return TCL_OK;
}
